import numpy as np
import sympy as sp
from sympy import symbols, Eq, solve

from GenPhylo.model.model_utils import alpha


def get_M1(distribution, dir_constant, exp_minus_l, sq_det_D):
 
    res = 1
    iteration = 1

    # Compute M1
    while res >= 1 and iteration < 50:
        M1 = np.zeros((4,4))
        i=0
        while i<4:
            dir = np.ones(4)
            dir[i] = dir_constant
            R = np.random.dirichlet(dir)
            if R[i] > 0.3:
                M1[i,:] = R
                i = i + 1

        new_distribution = np.matmul(distribution,M1)
        D_ = np.diag(new_distribution)
        sq_det_D_ = np.sqrt(np.linalg.det(D_))
        res = exp_minus_l * sq_det_D_ / sq_det_D
        detM1 = np.linalg.det(M1)

        #If detM1 > res and res < 1, conditions are met. Otherwise, a new iteration.
        if detM1 <= res:
            res = 1
        
        iteration += 1
    
    return M1, detM1, res, new_distribution


def get_M2(new_distribution,d2,dir_constant):
    """
    Metropolis - Hastings implementation to get M2
    """

    P = np.zeros((4,4))
    iteration = 0
    iter = True

    while iter and iteration < 50:
        # Random Markov matrix generation
        Q = np.zeros((4,4))
        i=0
        while i<4:
            dir = np.ones(4)
            dir[i] = dir_constant
            R = np.random.dirichlet(dir)
            if R[i] > 0.3:
                Q[i,:] = R
                i = i + 1

        # Time reversible matrix generation
        for i in range(4):
            for j in range(4):
                if i == j:
                    sum = 0
                    for k in range(4):
                        if k != i:
                            sum += (Q[i,k] * (1 - alpha(new_distribution,Q,i,k)))
                    P[i,j] = Q[i,i] + sum
                else:
                    P[i,j] = Q[i,j]*alpha(new_distribution,Q,i,j)

        assert (np.abs(np.sum(new_distribution - np.matmul(new_distribution,P)))) < 10**-6
        
        # Adjust the matrix diagonalising (ensure matrix with determinant d2)
        vaps, _ = np.linalg.eig(P)
        vaps = sorted(vaps, reverse=True)
        A = symbols('A')
        eq = Eq(-d2+(((1-A)*vaps[1]+A)*((1-A)*vaps[2]+A)*((1-A)*vaps[3]+A)),0)
        sol = solve(eq, A)
        # We only want the real solution between 0 and 1
        for s in sol:
            if s.is_real and 0 <= s <= 1:
                a = np.float64(s)
                M2 = (1-a)*P + a*np.identity(4)
                iter = False
                break
            elif s.is_complex: #If imaginary part is negligible
                if np.abs(np.imag(s)) < 10**-9 and 0 <= sp.re(s) <= 1:
                    a = np.float64(sp.re(s))
                    M2 = (1-a)*P + a*np.identity(4)
                    iter = False
                    break
        iteration += 1
    
    if iteration == 50:
        return 0
    else: 
        return M2


import numpy as np


def alpha(new_distribution, Q, i, k):
    """
    Returns the parameter alpha of the Metropolis - Hastings algorithm
    """
    ratio = new_distribution[k] * Q[k, i] / (new_distribution[i] * Q[i, k])
    return min(1, ratio)


def find_k(distribution, l, sq_det_D, exp_minus_l):
    """
    Finds a suitable value of k to satisfy the condition
    detM1 > np.exp(-l)*np.sqrt(np.linalg.det(D_))/np.sqrt(np.linalg.det(D)).
    """
    epsilon = 1e-3  # Desired precision
    lower_bound = 2.5
    upper_bound = 25  # Adjust upper bound as needed
    
    while upper_bound - lower_bound > epsilon:
        M1 = np.zeros((4,4))
        mid = (lower_bound + upper_bound) / 2
        dir_constant_mid = (mid*np.exp(-l/4))/(np.sqrt(l/4))
        i = 0
        while i<4:
            dir = np.ones(4)
            dir[i] = dir_constant_mid
            R = np.random.dirichlet(dir)
            if R[i] > 0.3:
                M1[i,:] = R
                i = i + 1
        
        new_distribution = np.matmul(distribution, M1)
        D_ = np.diag(new_distribution)
        detM1 = np.linalg.det(M1)
        sq_det_D_ = np.sqrt(np.linalg.det(D_))
        res = exp_minus_l * sq_det_D_ / sq_det_D

        if detM1 > res:
            upper_bound = mid
        else:
            lower_bound = mid

    return (lower_bound + upper_bound) / 2


def DLC(matrix):
    B = True
    for j in range(matrix.shape[1]):
        max_index = np.argmax(matrix[:, j])
        if max_index != j:
            B = False
    return B


def compare_equal_matrices(matrix1, matrix2):
    """
    Compares two matrices element-wise.
    """
    for i in range(matrix1.shape[0]):
        for j in range(matrix1.shape[1]):
            if matrix1[i, j] != matrix2[i, j]:
                return False
    return True


import numpy as np

from GenPhylo.model.matrix_computation import get_M1, get_M2
from GenPhylo.model.model_utils import find_k


def generate_random_matrix(distribution, l):
    """
    Returns the transition matrix M=M1M2 given a branch length
    and the distribution at the ancestor node.
    """

    D = np.diag(distribution)
    sq_det_D = np.sqrt(np.linalg.det(D))
    exp_minus_l = np.exp(-l)
    k = find_k(distribution, l, sq_det_D, exp_minus_l)
    dir_constant = (k*np.exp(-l/4))/(np.sqrt(l/4))

    M1, detM1, res, new_distribution = get_M1(distribution, dir_constant, exp_minus_l, sq_det_D)
    d2 = res * (1 / detM1)
    M2 = get_M2(new_distribution,d2,dir_constant)

    if not isinstance(M2, np.ndarray) and M2 == 0:
        return np.zeros((4,4))
    else:
        detM2 = np.linalg.det(M2)
        assert(np.abs(detM2 - d2) < 10**-6)
        M = np.matmul(M1,M2)
        return M




import numpy as np
import os
import tarfile


def generate_descendant_sequence(M, seq):
    """
    Given the sequence seq of the ancestor node and the transition matrix M,
    returns the sequence of the descendant node
    """
    nucleotide_map = {'A': 0, 'G': 1, 'C': 2, 'T': 3}
    row_indices = np.array([nucleotide_map[s] for s in seq])
    new_seq = ''.join(np.random.choice(['A', 'G', 'C', 'T'], p=M[row_indices[i], :]) for i in range(len(seq)))
    return new_seq


def generate_sequence(length, distribution):
    """
    Generates a sequence of length `length` using a given `distribution`.
    """
    nucleotides = np.array(['A', 'G', 'C', 'T'])
    seq = np.random.choice(nucleotides, size=length, p=distribution)
    return ''.join(seq)


def generate_alignments(net, node_distribution, edges, length, t, case, lengths, name):
    """
    Generates alignments based on the given parameters and saves them in FASTA files.
    """
    file_names = []
    tar_file = "./output_files/" + name + "_alignments.tar.gz"
    tar = tarfile.open(tar_file, "w:gz")

    if case == 1:
        for align in range(t):
            node_sequence = dict()
            node_sequence["Root"] = generate_sequence(length, node_distribution["Root"])
            i = 0
            for edge in net.edges():
                node_sequence[edge[1].name] = generate_descendant_sequence(edges[i].transition_matrix,
                                                                 node_sequence[edge[0].name])
                i += 1
            leaves_seq = {k: v for k, v in node_sequence.items() if k.startswith('L')}
            sequences_in_leaves = list(leaves_seq.values())
            keys_for_sequences = list(leaves_seq.keys())
            iter = 0
            file_name = str(len(sequences_in_leaves)) + "_leaves_" + str(
                len(sequences_in_leaves[0])) + "length_sequences_num" + str(
                align + 1) + ".fasta"
            file_names.append(file_name)
            file = open(file_name, "w")
            for seq in sequences_in_leaves:
                file.write(">Seq" + str(keys_for_sequences[iter]) + "\n" + seq + "\n")
                iter += 1
            file.close()
            tar.add(file_name)
            os.remove(file_name)
            
    else:
        for l in lengths:
            node_sequence = dict()
            node_sequence["Root"] = generate_sequence(l, node_distribution["Root"])
            i = 0
            for edge in net.edges():
                node_sequence[edge[1].name] = generate_descendant_sequence(edges[i].transition_matrix,
                                                                 node_sequence[edge[0].name])
                i += 1
            leaves_seq = {k: v for k, v in node_sequence.items() if k.startswith('L')}
            sequences_in_leaves = list(leaves_seq.values())
            keys_for_sequences = list(leaves_seq.keys())
            iter = 0
            file_name = str(len(sequences_in_leaves)) + "_leaves_" + str(
                len(sequences_in_leaves[0])) + "length_sequences.fasta"
            file_names.append(file_name)
            file = open(file_name, "w")
            for seq in sequences_in_leaves:
                file.write(">Seq" + str(keys_for_sequences[iter]) + "\n" + seq + "\n")
                iter += 1
            file.close()
            tar.add(file_name)
            os.remove(file_name)
    tar.close()
    return file_names


import os

def save_transition_matrices(matrices, name, newick_with_labels_str):
    """
    Saves the transition matrices and prints them.
    """
    real_matrices = []
    output_dir = './output_files'
    # Check if the directory exists, if not, create it
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    output_file = "./output_files/" + name + "_transition_matrices.txt"
    with open(output_file, "w") as f:
        f.write(newick_with_labels_str + "\n")
        for m in matrices:
            f.write(m.edge[0].name + " " + m.edge[1].name + "\n")
            f.write(str(m.transition_matrix) + "\n")
            f.write("**********************************************************************\n")
    return real_matrices